def simple_download(url, local_filename= None):
    """
    Input: 
    url: url to download
    local_filename: optional, by default will use the downloads folder
    from options. (Can be set with  
    Checks if a file exists at a filepath, and if not, requests the file.
    Note: If the directory does not exist, it will try to make one using os.mkdir
    so targeting a subdirectory is not supported (yet)
    Returns: the relative path to the file.
    """
    
    import core.options as opts
    
    import os
    import requests
    import time
    
    if not isinstance(url, str):
        raise TypeError("url must be str, not {}".format(
            type(url).__name__
        ))
    
    if not os.path.exists(opts.downloads_folder):
        os.mkdir(opts.downloads_folder)
    if not local_filename:
        local_filename = str(opts.downloads_folder + url.split('/')[-1]) #TODO: Improve
    if not os.path.exists(local_filename):
        r = requests.get(url)

        if r.status_code != 200: #Check connection
            raise ConnectionError(
                "Connection Failed")

        with open(local_filename, 'wb') as f:
            f.write(r.content)
    return local_filename
        

    

def down_extract_zip(url, target_ext, local_filename=None, take_first=True):
    """Downloads and unzips a shape file and returns
    the filepath of the file(s) as a string or list of strings
    
    Usage: flname = down_extract_zip(url_of_zipfile, 'shp', )
    flname is now the path of the shpfile.
    
    Inputs: 
    URL, a URL string
    target_ext, a string such as 'shp' or 'json'
    local_filename, the name of the file where the data 
    will be written
    take_first, default True, if False return a list of matching files
    in the unzipped shape file, if True return just the filepath
    as a string
    
    """
    
    import core.options as opts
    
    import zipfile
    import requests
    import time
    
    if not isinstance(url, str):
        raise TypeError("url must be str, not {}".format(
            type(url).__name__
        ))
    
    if not isinstance(target_ext, str):
        raise TypeError("target_ext must be str, not {}".format(
            type(target_ext).__name__
        ))
        
    if not isinstance(take_first, bool):
        raise TypeError("take_first must be bool, not {}".format(
            type(take_first).__name__
        ))
        
    if not local_filename: 
        local_filename = str(opts.downloads_folder + url.split('/')[-1])
    zip_fld = local_filename[:-4]
    if not os.path.exists(zip_fld):
        r = requests.get(url)

        if r.status_code != 200: #Check connection
            raise ConnectionError(
                "Connection Failed")

        with open(local_filename, 'wb') as f:
            f.write(r.content)
        os.makedirs(zip_fld)
        
        with zipfile.ZipFile(local_filename, "r") as zip_pl:
            zip_pl.extractall(zip_fld)

    fld_arr = os.listdir(zip_fld) #get array of unzipped files
    flname = None
    
    if take_first:
        #Find the first match
        
        for unzippedfile in fld_arr[::-1]:  #Hacky, replace later?
            if str(unzippedfile[-3:]) == target_ext:
                flname = zip_fld + '/' + unzippedfile
        
        if not flname: 
            raise FileNotFoundError('File with extension "{}" not found!'.format(target_ext))
        
        return flname
    
    if not take_first:
        #Return a list of filepaths
        
        matches = [path for path in fld_arr if path[-3:] == target_ext] 
        
        if len(matches) < 1:
            raise FileNotFoundError(
                'File with extension "{}" not found!'.format(
                    target_ext))
        
        if len(matches) >= 1:
            return matches


wrangler                 = 'pandas'


filepath                 = './data/'

#Filepath used by downloading_funcs.py
downloads_folder         = './data/'

#List of tuples in the form (callable, dict)
schedule                 = []

#Chunking indexes
chunk                    = []
chunkstype               = None
pandas_to_data           = None

#Threads timeout? arg of multiprocessing.join
timeouts                 = None #None or int

#Verbose execution?
loud                     = False


from __future__ import division
import sys
import os
import multiprocessing
from multiprocessing import Pipe


try:
    from core import *
except Exception as e:
    
    pass

import core.prefab
import core.options as opts
import core.download

class muni_data_pipe:
    """TODO: Docstring"""

    
    def __init__(self, wrangler='pandas', load='pandas', gis='geopandas', **kwargs):
                
        self.data     = None
        self.opts     = opts
        self.load     = load
        self.prefab   = prefab
        self.download = download
        
        
        #Default to Pandas if Dask won't wrangle
        if wrangler is 'pandas':
            import pandas as wg
        if wrangler is 'dask':
            try: 
                import dask as wg
            except ImportError:
                try:
                    import pandas as wg
                except ImportError:
                    raise ImportError(
            'Unable to import dask or pandas! (This should never happen.)'
                                     )
        self.wg = wg
        
        if self.load in ['SQLite']:
            sql_conn  = sqlite3.connect(self.opts.db_pth)
            curs     = sql_conn.cursor()
            curs.execute(self.opts.create_tables)
            
            
        if self.load == 'pandas':
            self.data = self.wg.DataFrame()
    
    def scheduler(self, chunk = None, sched = None):
        """TODO: Docstring"""
        
        if not sched:
            sched = self.opts.schedule
        
        if not chunk:
            chunk = self.opts.chunk
        
        
        sched_copy = [i for i in sched]
        sched_copy.reverse()
        
        while len(sched_copy) > 0:
            parent_conn, child_conn = Pipe()
            
            scheduled_func, message = sched_copy.pop()
            
            #The user must define chunk somewhere
            if 'chunk' in message:
                pass
            elif chunk:
                message['chunk'] = chunk
            else:
                raise ValueError(
            "The scheduler needs to know which chunk of data to use."
             )

            #Spawn the process and pass the connection, data and message to it
            if message:
                if 'load' in message:
                    if message['load'] == False:
                        p = multiprocessing.Process(target=scheduled_func, 
                                        args=(child_conn, None, message), 
                                        name = scheduled_func.__name__)
                        p.start()
                else:
                    p = multiprocessing.Process(target=scheduled_func, 
                                        args=(child_conn, self.data, message), 
                                        name = scheduled_func.__name__)
                    p.start()
            else:
                p = multiprocessing.Process(target=scheduled_func, 
                                        args=(child_conn, None, None), 
                                        name = scheduled_func.__name__)
                p.start()
                
             
                
            
            
            
            if self.opts.loud:
                print(
                    'Beginning next function: {}, {} remaining'.format(
                        p.name, len(sched_copy)))
            
            recv = parent_conn.recv()    
            p.join(timeout = self.opts.timeouts)
            
            if "BREAK" in recv:
                print("\n \nERROR IN CHILD PROCESS:\n")               
                print(recv["BREAK"])
                print(recv['TB'])
                raise recv["BREAK"]
                
                
            
            #Decide how to store result with Pandas
            if self.load == 'pandas':
                if "pandas_to_data" in message:
                    if message['pandas_to_data'] in [
                        'append', 'concat', 'overwrite', 'none',]:
                        pass
                    
                elif self.opts.pandas_to_data in [
                    'append', 'concat', 'overwrite', 'none',]:
                    
                    message['pandas_to_data'] = self.opts.pandas_to_data
                else:
                    raise ValueError(
        "To store data with pandas you must define "             
        "the pandas_to_data option or pass it to your UDF "
        "\nSupports: 'none', 'append', 'concat', 'merge' and 'overwrite'"
                                    )
                
                if message['pandas_to_data'] == 'none':
                    pass
                if message['pandas_to_data'] == 'append':
                    if "pandas_to_data_args" in message:
                        pandas_to_data_args = message['pandas_to_data_args']
                        self.data.append(recv,
                                         **pandas_to_data_args)
                    else:
                        self.data.append(recv)
                if message['pandas_to_data'] == 'concat':
                    if "pandas_to_data_args" in message:
                        pandas_to_data_args = message['pandas_to_data_args']
                        self.data.concat(recv,
                                         **pandas_to_data_args)
                    else:
                        self.data.concat(recv)
                if message['pandas_to_data'] == 'merge':
                    if "pandas_to_data_args" in message:
                        pandas_to_data_args = message['pandas_to_data_args']
                        self.data.append(recv,
                                         **pandas_to_data_args)
                    else:
                        raise ValueError(
                        "You must define pandas_to_data_args "
                            "with the merge option"
                        )
                if message['pandas_to_data'] == 'overwrite':
                    self.data = recv
            
           
            
            #Not yet implemented
            if self.load == 'postgreSQL':
                raise ValueError("postgreSQL is not currently supported")
            
            #Not yet implemented
            if self.load == 'SQLite':
                raise ValueError("SQLite is not currently supported")

            
            
            #Join the process and move on
            
            
            if self.opts.loud:
                print(
                    'Scheduled function completed, {} remaining.'.format(
                        len(sched_copy)))
    
    #TODO: build this out if needed
    def _stack_local_data(self, filepath, filetype=None):
        
        if not os.path.exists(filepath):
            raise FileNotFoundError(
            "File {} does not exist.".format(filepath)
            )
        
        if self.wg.__name__ == 'pandas':
            pass
        
        if self.wg.__name__ == 'dask':
            pass
    
    
    def chunker(self, chunks=None, chunk_args=None):
        """Method for telling the scheduler which data to use
        up the data."""
        
        
        
        #Ensure chunks is assigned and valid
        if not chunks:
            if type(self.opts.chunk).__name__ in ['list', 'tuple']:
                chunks = [i for i in self.opts.chunk]
            elif type(self.opts.chunk).__name__ == 'str':
                chunks = self.opts.chunk
        if type(chunks).__name__ not in ['list', 'tuple', 'str']: 
            raise ValueError(
                "chunks must be list, tuple or str, not {}".format(
                type(chunks).__name__))

        for i in range(1 + len(chunks)):
            if self.opts.loud == True:
                print("\nMoving to next chunk: "
                      "\n on chunk number {}\n".format(i))    
            self.scheduler(chunk=chunks[i])
            
        


#!/usr/bin/env python

from setuptools import setup

setup(name='munidatapipeline',
      version='00.02.01',
      description='An ETL tool for working with municipal open data (especially GiS data) on resource constrained systems.',
      author='Charles Landau, Andrew Gobbi',
      author_email='munidatapipelinedevelopment@gmail.com',
      url='https://github.com/mdp-dev/munidatapipeline',
      keywords=['munidatapipeline', 'ETL', 'multiprocessing'],
      packages=['munidatapipeline', 'core'],
     )



import sys
import os
import traceback
from shapely.geometry import Point
import core.download as dlf
import pandas as pd
import geopandas as gpd


def err_to_parent(UDF):

    def handling(connection, load, message):
        try:
            UDF(connection, load, message)
        except Exception as e:
            
            connection.send({"BREAK": e,
                            "TB": traceback.format_exc()})
    handling.__name__ = UDF.__name__
    return handling

@err_to_parent
def simple_groupby(connection, load, message):
    """
    Return the result of a Pandas groupby method for
    the load dataframe.
    
    Inputs: 
    
    load (must be DataFrame or GeoDataFrame)
    message['col_name'], the series to group by
    message['groupby_method'], may be count, mean, sum, or median
    
    Output: sends the result of the method back up the pipe
    
    """
    
    if 'groupby_method' in message:
        groupby_method = message['groupby_method']
    else:
        raise ValueError('groupby_method must be sum, count, mean, or median')
    
    if 'col_name' not in message:
        raise ValueError("message['col_name'] not defined")
    
    if type(load).__name__ not in ['DataFrame', 'GeoDataFrame']:
        raise TypeError('load must be DataFrame or GeoDataFrame, not {}'.format(
            type(load).__name__))
                        
    
    valid_methods = ['sum', 'count', 'mean', 'median']
    if groupby_method not in(valid_methods):
        raise ValueError(
            'groupby must be sum, count, mean, or median , not {}'.format(
                groupby_method))
    if groupby_method == 'sum':
           grouped = load.groupby(message['col_name']).sum()
    if groupby_method == 'count':
           grouped = load.groupby(message['col_name']).count()
    if groupby_method == 'mean':
           grouped = load.groupby(message['col_name']).mean()
    if groupby_method == 'median':
           grouped = load.groupby(message['col_name']).median()
    
    connection.send(grouped)

@err_to_parent    
def points_from_latlong(connection, load, message):
    """
    Render a lat and long column to shapely points.
    
    Inputs: connection, load (a DataFrame or GeoDataFrame), message (see below)
    
        Message:
            message['lat'], the key/column name for lattitude listlike
            message['long'], the key/column name for the longitude listlike
            message['out_column'], name of new points column to be created

    Output: sends the load back up the pipe with a new points column
    
    """
    
    
    
    long = message['long']
    lat  = message['lat']
    
    load[message['out_column']] = [Point(x, y) for x, y in zip(
        load[long], load[lat]
    )]
    
    connection.send(load)

@err_to_parent    
def col_to_datetime (connection, load=None, message=None): 
    #turn list of columns into a date time object
    
    if type(load).__name__ == 'DataFrame':

        if not isinstance(message['col_to_datetime'], list):                 
                raise TypeError("message['col_to_datetime'] must be a list")
        if message['col_to_datetime'] == []:
                raise ValueError(
                    'List of columns to be converted to datetime object is empty')

        for i in message['col_to_datetime']:
            load[i] = pd.to_datetime(i)
        connection.send(load)

@err_to_parent
def data_down(connection, load, message):
    #A simple way to use the scheduler for downloading data

    if 'urls' not in message:
        raise ValueError("Required value 'message['urls']' missing!")
    
    for i in message['urls']:
        dlf.simple_download(i)
    
    print("Sending filepath!")
    connection.send("Finished Downloading")
        
        

        
@err_to_parent        
def pd_simple_merge(connection, load, message):
    """
    Perform a merge on two dataframes with the pandas merge.
    
    The left DataFrame is always the load argument while the
    right DataFrame is always constructed from a file.
    
    Requires: Pandas
    Inputs: connection, load (the left side, 
    a DataFrame), and message(see below)
    
        Message:
            message['path'], filepath for the right side of the join
            
            message['merge_args'], dict of required args for joining 
            the DataFrames
            
            message['DataFrame_right_args'], dict of 
            required args for constructing the DataFrame 
            from the path using the pd.DataFrame command
            
    
    Output: send the merged load back up the pipe
    
    """
    
    if type(load).__name__ != 'DataFrame':
        raise ValueError("load must be a Pandas DataFrame not {}".format(
            type(load).__name__))
    
    left  = load
    right = pd.DataFrame(message['path'], **message['DataFrame_right_args'])
    
    merged = pd.merge(left, right, **message['merge_args'])
    
@err_to_parent    
def pandas_frame_file(connection, load, message):
    """Simply read a file into a dataframe and send it back up the pipe.
    Input: filepath, filetype, read_args in message.
    Note: If message['use_chunk'] is True, then the function
    takes the filepath from the scheduler's current chunk
    
    Output: Sends a dataframe back up the pipe."""
    
    
    if 'use_chunk' in message:
        if message['use_chunk']:
            filepath = message['chunk']
    elif 'filepath' in message:
         filepath = message['filepath']
    
    try:
        os.path.exists(filepath)
    except Exception as e:
        raise e
    
    #Pick a reader and send back the df
    df = _pd_reader_pick(message)
    connection.send(df)

@err_to_parent
def simple_gpd_sjoin(connection, load, message):
    """
    Perform an sjoin on two dataframes with GeoPandas.
    
    If the load is not of type GeoDataFrame then try to
    construct it into one.
    
    The other GeoDataFrame is always constructed from a file.
    
    Requires: Geopandas
    Inputs: connection, load (the left side, 
    a DataFrame or GeoDataFrame), and message(see below)
    
        Message:
            message['path'], filepath for the right side of the join
            
            message['sjoin_args'], required args for joining 
            the GeoDataFrames
            
            message['GeoDataFrame_right_args'], required args for constructing
            the GeoDataFrame from the path using the gpd.read_file command
            
            message['GeoDataFrame_left_args'], optional args for constructing 
            the GeoDataFrame from the load using the gpd.read_file command
    
    Output: send the sjoined load back up the pipe
    """

    
    for i in ['sjoin_args', 'GeoDataFrame_right_args', 'path']:
        if i not in message:
            raise KeyError("{}".format(i))
    
    if type(load).__name__ != 'GeoDataFrame':
        if 'GeoDataFrame_left_args' in message:
            left = gpd.GeoDataFrame(load, **message['GeoDataFrame_left_args'])
        else:
            raise ValueError('GeoDataFrame_left_args is required if the load'
                            ' is not already a GeoDataFrame')
    else:
        left = load
    


    right = gpd.read_file(message['right_path'],
                **message['GeoDataFrame_right_args'])
    
    
    
    
@err_to_parent    
def pd_simple_merge(connection, load, message):
    """
    Perform a merge on two dataframes with the pandas merge.
    
    The left DataFrame is always the load argument while the
    right DataFrame is always constructed from a file.
    
    Requires: Pandas
    Inputs: connection, load (the left side, 
    a DataFrame), and message(see below)
    
        Message:
            message['path'], filepath for the right side of the join
            
            message['merge_args'], required args for joining 
            the DataFrames
            
            message['DataFrame_right_args'], required args for constructing
            the DataFrame from the path using the gpd.read_file command
            
    
    Output: send the merged load back up the pipe
    
    """
    

    
    if type(load).__name__ != 'DataFrame':
        raise ValueError("load must be a Pandas DataFrame not {}".format(
            type(load).__name__))
    
    left  = load
    right = pd.DataFrame(message['path'], **message['DataFrame_right_args'])
    
    merged = pd.merge(left, right, **message['merge_args'])
    
    connection.send(merged)
    
def _pd_reader_pick(message):
    
    if 'use_chunk' in message:
        if message['use_chunk']:
            filepath = message['chunk']
    elif not 'filepath' or 'use_chunk' in message:
        raise ValueError("Must pass message['filepath'] or message['use_chunk']")
    else:
        filepath = message['filepath']
    
    if not 'filetype' in message:
        raise ValueError("Must pass message['filetype'] \n"
                        "Supported types: 'csv', 'excel', 'pkl'")
    
    
    #Construct dataframes (with args if neccessary)
    if message['filetype'] == "csv":
        if 'read_args' in message:
            df = pd.read_csv(filepath, **message['read_args'])
        else:
            df = pd.read_csv(filepath)
    if message['filetype'] == "pkl":
        if 'read_args' in message:
            df = pd.read_pickle(filepath, **message['read_args'])
        else:
            df = pd.read_pickle(filepath)
    if message['filetype'] == "excel":
        if 'read_args' in message:
            df = pd.read_excel(filepath, **message['read_args'])
        else:
            df = pd.read_excel(filepath)
    return df

def pd_writer(connection, load, message):
    """Exposes the pandas to_file operations.
    Inputs:
    
    load, the dataframe to be written
    message['filetype'], can be 'csv', 'hd5' or 'excel'
    message['pd_args'], args for the pandas file writer
    message['path'], optional, filepath to write to, overrides opts.filepath
        
    Outputs: writes a file to target path
    """
    
    